"""
YAML schemas.

Contents:

* ``config_file_schema.yaml`` – Schema for configuration selection files.
* ``config_schema.yaml`` – Schema for configuration files.
* ``environments_spec_schema.yaml`` – Schema for execution environment definition files.
* ``files_spec_schema.yaml`` – Schema for input/output specification files.
* ``parameters_spec_schema.yaml`` – Schema for parameter specification files.
* ``task_schema_spec_schema.yaml`` – Schema for task template specification files.
* ``workflow_spec_schema.yaml`` – Schema for workflow files.
"""
